import React, { Component } from "react"

function shallowEqualArrays(arrA, arrB) {
  if (arrA === arrB) {
    return true
  }

  let len = arrA.length

  if (arrB.length !== len) {
    return false
  }

  for (let i = 0; i < len; i++) {
    if (arrA[i] !== arrB[i]) {
      return false
    }
  }

  return true
}

function createSectionIterator({ data, multiSection }) {
  function nextNonEmptySectionIndex(sectionIndex) {
    if (sectionIndex === null) {
      sectionIndex = 0
    } else {
      sectionIndex++
    }

    while (sectionIndex < data.length && data[sectionIndex] === 0) {
      sectionIndex++
    }

    return sectionIndex === data.length ? null : sectionIndex
  }

  function prevNonEmptySectionIndex(sectionIndex) {
    if (sectionIndex === null) {
      sectionIndex = data.length - 1
    } else {
      sectionIndex--
    }

    while (sectionIndex >= 0 && data[sectionIndex] === 0) {
      sectionIndex--
    }

    return sectionIndex === -1 ? null : sectionIndex
  }

  function next(position) {
    let [sectionIndex, itemIndex] = position

    if (multiSection) {
      if (itemIndex === null || itemIndex === data[sectionIndex] - 1) {
        sectionIndex = nextNonEmptySectionIndex(sectionIndex)

        if (sectionIndex === null) {
          return [null, null]
        }

        return [sectionIndex, 0]
      }

      return [sectionIndex, itemIndex + 1]
    }

    if (data === 0 || itemIndex === data - 1) {
      return [null, null]
    }

    if (itemIndex === null) {
      return [null, 0]
    }

    return [null, itemIndex + 1]
  }

  function prev(position) {
    let [sectionIndex, itemIndex] = position

    if (multiSection) {
      if (itemIndex === null || itemIndex === 0) {
        sectionIndex = prevNonEmptySectionIndex(sectionIndex)

        if (sectionIndex === null) {
          return [null, null]
        }

        return [sectionIndex, data[sectionIndex] - 1]
      }

      return [sectionIndex, itemIndex - 1]
    }

    if (data === 0 || itemIndex === 0) {
      return [null, null]
    }

    if (itemIndex === null) {
      return [null, data - 1]
    }

    return [null, itemIndex - 1]
  }

  function isLast(position) {
    return next(position)[1] === null
  }

  return {
    next,
    prev,
    isLast
  }
}

function compareObjects(objA, objB, keys = []) {
  if (objA === objB) {
    return false
  }

  const aKeys = Object.keys(objA)
  const bKeys = Object.keys(objB)

  if (aKeys.length !== bKeys.length) {
    return true
  }

  const keysMap = {}
  let i, len

  for ((i = 0), (len = keys.length); i < len; i++) {
    keysMap[keys[i]] = true
  }

  for ((i = 0), (len = aKeys.length); i < len; i++) {
    let key = aKeys[i]
    const aValue = objA[key]
    const bValue = objB[key]

    if (aValue === bValue) {
      continue
    }

    if (
      !keysMap[key] ||
      aValue === null ||
      bValue === null ||
      typeof aValue !== "object" ||
      typeof bValue !== "object"
    ) {
      return true
    }

    const aValueKeys = Object.keys(aValue)
    const bValueKeys = Object.keys(bValue)

    if (aValueKeys.length !== bValueKeys.length) {
      return true
    }

    for (let n = 0, length = aValueKeys.length; n < length; n++) {
      const aValueKey = aValueKeys[n]

      if (aValue[aValueKey] !== bValue[aValueKey]) {
        return true
      }
    }
  }

  return false
}

function themeable(input) {
  const [theme, classNameDecorator] = Array.isArray(input) && input.length === 2
    ? input
    : [input, null]

  return (key, ...names) => {
    const styles = names.map(name => theme[name]).filter(x => x)

    return typeof styles[0] === "string" ||
      typeof classNameDecorator === "function"
      ? {
          key,
          className: classNameDecorator
            ? classNameDecorator(...styles)
            : styles.join(" ")
        }
      : { key, style: Object.assign({}, ...styles) }
  }
}

const alwaysTrue = () => true

const defaultTheme = {
  container: "react-autosuggest__container",
  containerOpen: "react-autosuggest__container--open",
  input: "react-autosuggest__input",
  itemsContainer: "react-autosuggest__suggestions-container",
  itemsList: "react-autosuggest__suggestions-list",
  item: "react-autosuggest__suggestion",
  itemFocused: "react-autosuggest__suggestion--focused",
  sectionContainer: "react-autosuggest__section-container",
  sectionTitle: "react-autosuggest__section-title"
}

class SectionTitle extends Component {
  shouldComponentUpdate(nextProps) {
    return compareObjects(nextProps, this.props)
  }

  render() {
    const { section, renderSectionTitle, theme, sectionKeyPrefix } = this.props
    const sectionTitle = renderSectionTitle(section)

    if (!sectionTitle) {
      return null
    }

    return (
      <div {...theme(`${sectionKeyPrefix}title`, "sectionTitle")}>
        {sectionTitle}
      </div>
    )
  }
}

class Item extends Component {
  constructor() {
    super()

    this.storeItemReference = this.storeItemReference.bind(this)
    this.onMouseEnter = this.onMouseEnter.bind(this)
    this.onMouseLeave = this.onMouseLeave.bind(this)
    this.onMouseDown = this.onMouseDown.bind(this)
    this.onClick = this.onClick.bind(this)
  }

  shouldComponentUpdate(nextProps) {
    return compareObjects(nextProps, this.props, ["renderItemData"])
  }

  storeItemReference(item) {
    if (item !== null) {
      this.item = item
    }
  }

  onMouseEnter(event) {
    const { sectionIndex, itemIndex } = this.props

    this.props.onMouseEnter(event, { sectionIndex, itemIndex })
  }

  onMouseLeave(event) {
    const { sectionIndex, itemIndex } = this.props

    this.props.onMouseLeave(event, { sectionIndex, itemIndex })
  }

  onMouseDown(event) {
    const { sectionIndex, itemIndex } = this.props

    this.props.onMouseDown(event, { sectionIndex, itemIndex })
  }

  onClick(event) {
    const { sectionIndex, itemIndex } = this.props

    this.props.onClick(event, { sectionIndex, itemIndex })
  }

  render() {
    const { item, renderItem, renderItemData, ...restProps } = this.props

    delete restProps.sectionIndex
    delete restProps.itemIndex

    if (typeof restProps.onMouseEnter === "function") {
      restProps.onMouseEnter = this.onMouseEnter
    }

    if (typeof restProps.onMouseLeave === "function") {
      restProps.onMouseLeave = this.onMouseLeave
    }

    if (typeof restProps.onMouseDown === "function") {
      restProps.onMouseDown = this.onMouseDown
    }

    if (typeof restProps.onClick === "function") {
      restProps.onClick = this.onClick
    }

    return (
      <li role="option" {...restProps} ref={this.storeItemReference}>
        {renderItem(item, renderItemData)}
      </li>
    )
  }
}

class ItemsList extends Component {
  static defaultProps = {
    sectionIndex: null
  }

  constructor() {
    super()

    this.storeFocusedItemReference = this.storeFocusedItemReference.bind(this)
  }

  shouldComponentUpdate(nextProps) {
    return compareObjects(nextProps, this.props, ["itemProps"])
  }

  storeFocusedItemReference(focusedItem) {
    this.props.onFocusedItemChange(
      focusedItem === null ? null : focusedItem.item
    )
  }

  render() {
    const {
      items,
      itemProps,
      renderItem,
      renderItemData,
      sectionIndex,
      focusedItemIndex,
      getItemId,
      theme,
      keyPrefix
    } = this.props
    const sectionPrefix = sectionIndex === null
      ? keyPrefix
      : `${keyPrefix}section-${sectionIndex}-`
    const isItemPropsFunction = typeof itemProps === "function"

    return (
      <ul role="listbox" {...theme(`${sectionPrefix}items-list`, "itemsList")}>
        {items.map((item, itemIndex) => {
          const isFocused = itemIndex === focusedItemIndex
          const itemKey = `${sectionPrefix}item-${itemIndex}`
          const itemPropsObj = isItemPropsFunction
            ? itemProps({ sectionIndex, itemIndex })
            : itemProps
          const allItemProps = {
            id: getItemId(sectionIndex, itemIndex),
            ...theme(itemKey, "item", isFocused && "itemFocused"),
            ...itemPropsObj
          }

          if (isFocused) {
            allItemProps.ref = this.storeFocusedItemReference
          }

          // `key` is provided by theme()
          /* eslint-disable react/jsx-key */
          return (
            <Item
              {...allItemProps}
              sectionIndex={sectionIndex}
              itemIndex={itemIndex}
              item={item}
              renderItem={renderItem}
              renderItemData={renderItemData}
            />
          )
          /* eslint-enable react/jsx-key */
        })}
      </ul>
    )
  }
}

class Autowhatever extends Component {
  static defaultProps = {
    id: "1",
    multiSection: false,
    renderInputComponent: props => <input {...props} />,
    renderItemsContainer: props => <div {...props} />,
    shouldRenderSection: alwaysTrue,
    renderItem: () => {
      throw new Error("`renderItem` must be provided")
    },
    renderItemData: {},
    renderSectionTitle: () => {
      throw new Error("`renderSectionTitle` must be provided")
    },
    getSectionItems: () => {
      throw new Error("`getSectionItems` must be provided")
    },
    inputProps: {},
    itemProps: {},
    focusedSectionIndex: null,
    focusedItemIndex: null,
    theme: defaultTheme
  }

  constructor(props) {
    super(props)

    this.focusedItem = null

    this.setSectionsItems(props)
    this.setSectionIterator(props)
    this.setTheme(props)

    this.onKeyDown = this.onKeyDown.bind(this)
    this.storeInputReference = this.storeInputReference.bind(this)
    this.storeItemsContainerReference = this.storeItemsContainerReference.bind(
      this
    )
    this.onFocusedItemChange = this.onFocusedItemChange.bind(this)
    this.getItemId = this.getItemId.bind(this)
  }

  componentDidMount() {
    this.ensureFocusedItemIsVisible()
  }

  componentWillReceiveProps(nextProps) {
    if (nextProps.items !== this.props.items) {
      this.setSectionsItems(nextProps)
    }

    if (
      nextProps.items !== this.props.items ||
      nextProps.multiSection !== this.props.multiSection
    ) {
      this.setSectionIterator(nextProps)
    }

    if (nextProps.theme !== this.props.theme) {
      this.setTheme(nextProps)
    }
  }

  componentDidUpdate() {
    this.ensureFocusedItemIsVisible()
  }

  setSectionsItems(props) {
    if (props.multiSection) {
      this.sectionsItems = props.items.map(section =>
        props.getSectionItems(section)
      )
      this.sectionsLengths = this.sectionsItems.map(items => items.length)
      this.allSectionsAreEmpty = this.sectionsLengths.every(
        itemsCount => itemsCount === 0
      )
    }
  }

  setSectionIterator(props) {
    this.sectionIterator = createSectionIterator({
      multiSection: props.multiSection,
      data: props.multiSection ? this.sectionsLengths : props.items.length
    })
  }

  setTheme(props) {
    this.theme = themeable(props.theme)
  }

  storeInputReference(input) {
    if (input !== null) {
      this.input = input
    }
  }

  storeItemsContainerReference(itemsContainer) {
    if (itemsContainer !== null) {
      this.itemsContainer = itemsContainer
    }
  }

  onFocusedItemChange(focusedItem) {
    this.focusedItem = focusedItem
  }

  getItemId(sectionIndex, itemIndex) {
    if (itemIndex === null) {
      return null
    }

    const { id } = this.props
    const section = sectionIndex === null ? "" : `section-${sectionIndex}`

    return `react-autowhatever-${id}-${section}-item-${itemIndex}`
  }

  renderSections() {
    if (this.allSectionsAreEmpty) {
      return null
    }

    const { theme } = this
    const {
      id,
      items,
      renderItem,
      renderItemData,
      shouldRenderSection,
      renderSectionTitle,
      focusedSectionIndex,
      focusedItemIndex,
      itemProps
    } = this.props

    return items.map((section, sectionIndex) => {
      if (!shouldRenderSection(section)) {
        return null
      }

      const keyPrefix = `react-autowhatever-${id}-`
      const sectionKeyPrefix = `${keyPrefix}section-${sectionIndex}-`

      // `key` is provided by theme()
      /* eslint-disable react/jsx-key */
      return (
        <div {...theme(`${sectionKeyPrefix}container`, "sectionContainer")}>
          <SectionTitle
            section={section}
            renderSectionTitle={renderSectionTitle}
            theme={theme}
            sectionKeyPrefix={sectionKeyPrefix}
          />
          <ItemsList
            items={this.sectionsItems[sectionIndex]}
            itemProps={itemProps}
            renderItem={renderItem}
            renderItemData={renderItemData}
            sectionIndex={sectionIndex}
            focusedItemIndex={
              focusedSectionIndex === sectionIndex ? focusedItemIndex : null
            }
            onFocusedItemChange={this.onFocusedItemChange}
            getItemId={this.getItemId}
            theme={theme}
            keyPrefix={keyPrefix}
            ref={this.storeItemsListReference}
          />
        </div>
      )
      /* eslint-enable react/jsx-key */
    })
  }

  renderItems() {
    const { items } = this.props

    if (items.length === 0) {
      return null
    }

    const { theme } = this
    const {
      id,
      renderItem,
      renderItemData,
      focusedSectionIndex,
      focusedItemIndex,
      itemProps
    } = this.props

    return (
      <ItemsList
        items={items}
        itemProps={itemProps}
        renderItem={renderItem}
        renderItemData={renderItemData}
        focusedItemIndex={
          focusedSectionIndex === null ? focusedItemIndex : null
        }
        onFocusedItemChange={this.onFocusedItemChange}
        getItemId={this.getItemId}
        theme={theme}
        keyPrefix={`react-autowhatever-${id}-`}
      />
    )
  }

  onKeyDown(event) {
    const { inputProps, focusedSectionIndex, focusedItemIndex } = this.props

    switch (event.key) {
      case "ArrowDown":
      case "ArrowUp": {
        const nextPrev = event.key === "ArrowDown" ? "next" : "prev"
        const [
          newFocusedSectionIndex,
          newFocusedItemIndex
        ] = this.sectionIterator[nextPrev]([
          focusedSectionIndex,
          focusedItemIndex
        ])

        inputProps.onKeyDown(event, {
          newFocusedSectionIndex,
          newFocusedItemIndex
        })
        break
      }

      default:
        inputProps.onKeyDown(event, { focusedSectionIndex, focusedItemIndex })
    }
  }

  ensureFocusedItemIsVisible() {
    const { focusedItem } = this

    if (!focusedItem) {
      return
    }

    const { itemsContainer } = this
    const itemOffsetRelativeToContainer = focusedItem.offsetParent ===
      itemsContainer
      ? focusedItem.offsetTop
      : focusedItem.offsetTop - itemsContainer.offsetTop

    let { scrollTop } = itemsContainer // Top of the visible area

    if (itemOffsetRelativeToContainer < scrollTop) {
      // Item is off the top of the visible area
      scrollTop = itemOffsetRelativeToContainer
    } else if (
      itemOffsetRelativeToContainer + focusedItem.offsetHeight >
      scrollTop + itemsContainer.offsetHeight
    ) {
      // Item is off the bottom of the visible area
      scrollTop =
        itemOffsetRelativeToContainer +
        focusedItem.offsetHeight -
        itemsContainer.offsetHeight
    }

    if (scrollTop !== itemsContainer.scrollTop) {
      itemsContainer.scrollTop = scrollTop
    }
  }

  render() {
    const { theme } = this
    const {
      id,
      multiSection,
      renderInputComponent,
      renderItemsContainer,
      focusedSectionIndex,
      focusedItemIndex
    } = this.props
    const renderedItems = multiSection
      ? this.renderSections()
      : this.renderItems()
    const isOpen = renderedItems !== null
    const ariaActivedescendant = this.getItemId(
      focusedSectionIndex,
      focusedItemIndex
    )
    const containerProps = theme(
      `react-autowhatever-${id}-container`,
      "container",
      isOpen && "containerOpen"
    )
    const itemsContainerId = `react-autowhatever-${id}`
    const inputComponent = renderInputComponent({
      type: "text",
      value: "",
      autoComplete: "off",
      role: "combobox",
      "aria-autocomplete": "list",
      "aria-owns": itemsContainerId,
      "aria-expanded": isOpen,
      "aria-haspopup": isOpen,
      "aria-activedescendant": ariaActivedescendant,
      ...theme(`react-autowhatever-${id}-input`, "input"),
      ...this.props.inputProps,
      onKeyDown: this.props.inputProps.onKeyDown && this.onKeyDown,
      ref: this.storeInputReference
    })
    const itemsContainer = renderItemsContainer({
      id: itemsContainerId,
      ...theme(`react-autowhatever-${id}-items-container`, "itemsContainer"),
      ref: this.storeItemsContainerReference,
      children: renderedItems
    })

    return (
      <div {...containerProps}>
        {inputComponent}
        {itemsContainer}
      </div>
    )
  }
}

export default class ReactimaSuggest extends Component {
  static defaultProps = {
    shouldRenderSuggestions: value => value.trim().length > 0,
    alwaysRenderSuggestions: false,
    multiSection: false,
    focusInputOnSuggestionClick: true,
    focusFirstSuggestion: false,
    theme: {
      container: "react-autosuggest__container",
      containerOpen: "react-autosuggest__container--open",
      input: "react-autosuggest__input",
      suggestionsContainer: "react-autosuggest__suggestions-container",
      suggestionsList: "react-autosuggest__suggestions-list",
      suggestion: "react-autosuggest__suggestion",
      suggestionFocused: "react-autosuggest__suggestion--focused",
      sectionContainer: "react-autosuggest__section-container",
      sectionTitle: "react-autosuggest__section-title"
    },
    id: "1"
  }

  constructor({ alwaysRenderSuggestions }) {
    super()

    this.state = {
      isFocused: false,
      isCollapsed: !alwaysRenderSuggestions,
      focusedSectionIndex: null,
      focusedSuggestionIndex: null,
      valueBeforeUpDown: null
    }

    this.justPressedUpDown = false
  }

  componentDidMount() {
    document.addEventListener("mousedown", this.onDocumentMouseDown)
  }

  componentWillReceiveProps(nextProps) {
    if (shallowEqualArrays(nextProps.suggestions, this.props.suggestions)) {
      if (
        nextProps.focusFirstSuggestion &&
        nextProps.suggestions.length > 0 &&
        this.justPressedUpDown === false
      ) {
        this.focusFirstSuggestion()
      }
    } else {
      if (this.willRenderSuggestions(nextProps)) {
        if (nextProps.focusFirstSuggestion) {
          this.focusFirstSuggestion()
        }

        if (this.state.isCollapsed && !this.justSelectedSuggestion) {
          this.revealSuggestions()
        }
      } else {
        this.resetFocusedSuggestion()
      }
    }
  }

  componentWillUnmount() {
    document.removeEventListener("mousedown", this.onDocumentMouseDown)
  }

  inputFocused(shouldRender) {
    this.setState({
      isFocused: true,
      isCollapsed: !shouldRender
    })
  }

  inputBlurred(shouldRender) {
    this.setState({
      isFocused: false,
      focusedSectionIndex: null,
      focusedSuggestionIndex: null,
      valueBeforeUpDown: null,
      isCollapsed: !shouldRender
    })
  }

  inputChanged(shouldRender) {
    this.setState({
      focusedSectionIndex: null,
      focusedSuggestionIndex: null,
      valueBeforeUpDown: null,
      isCollapsed: !shouldRender
    })
  }

  updateFocusedSuggestion(sectionIndex, suggestionIndex, prevValue) {
    let { valueBeforeUpDown } = this.state

    if (suggestionIndex === null) {
      valueBeforeUpDown = null
    } else if (valueBeforeUpDown === null && typeof prevValue !== "undefined") {
      valueBeforeUpDown = prevValue
    }

    this.setState({
      focusedSectionIndex: sectionIndex,
      focusedSuggestionIndex: suggestionIndex,
      valueBeforeUpDown
    })
  }

  resetFocusedSuggestion(shouldResetValueBeforeUpDown = true) {
    const { valueBeforeUpDown } = this.state

    this.setState({
      focusedSectionIndex: null,
      focusedSuggestionIndex: null,
      valueBeforeUpDown: shouldResetValueBeforeUpDown ? null : valueBeforeUpDown
    })
  }

  revealSuggestions() {
    this.setState({
      isCollapsed: false
    })
  }

  closeSuggestions() {
    this.setState({
      focusedSectionIndex: null,
      focusedSuggestionIndex: null,
      valueBeforeUpDown: null,
      isCollapsed: true
    })
  }

  getSuggestion(sectionIndex, suggestionIndex) {
    const { suggestions, multiSection, getSectionSuggestions } = this.props

    if (multiSection) {
      return getSectionSuggestions(suggestions[sectionIndex])[suggestionIndex]
    }

    return suggestions[suggestionIndex]
  }

  getFocusedSuggestion() {
    const { focusedSectionIndex, focusedSuggestionIndex } = this.state

    if (focusedSuggestionIndex === null) {
      return null
    }

    return this.getSuggestion(focusedSectionIndex, focusedSuggestionIndex)
  }

  getSuggestionValueByIndex(sectionIndex, suggestionIndex) {
    const { getSuggestionValue } = this.props

    return getSuggestionValue(this.getSuggestion(sectionIndex, suggestionIndex))
  }

  getSuggestionIndices(suggestionElement) {
    const sectionIndex = suggestionElement.getAttribute("data-section-index")
    const suggestionIndex = suggestionElement.getAttribute(
      "data-suggestion-index"
    )

    return {
      sectionIndex: typeof sectionIndex === "string"
        ? parseInt(sectionIndex, 10)
        : null,
      suggestionIndex: parseInt(suggestionIndex, 10)
    }
  }

  onDocumentMouseDown = event => {
    this.justClickedOnSuggestionsContainer = false

    let node =
      (event.detail && event.detail.target) || // This is for testing only. Please show me a better way to emulate this.
      event.target

    do {
      if (node.getAttribute("data-suggestion-index") !== null) {
        // Suggestion was clicked
        return
      }

      if (node === this.suggestionsContainer) {
        // Something else inside suggestions container was clicked
        this.justClickedOnSuggestionsContainer = true
        return
      }

      node = node.parentNode
    } while (node !== null && node !== document)
  }

  findSuggestionElement(startNode) {
    let node = startNode

    do {
      if (node.getAttribute("data-suggestion-index") !== null) {
        return node
      }

      node = node.parentNode
    } while (node !== null)

    console.error("Clicked element:", startNode) // eslint-disable-line no-console
    throw new Error("Couldn't find suggestion element")
  }

  maybeCallOnChange(event, newValue, method) {
    const { value, onChange } = this.props.inputProps

    if (newValue !== value) {
      onChange(event, { newValue, method })
    }
  }

  willRenderSuggestions(props) {
    const { suggestions, inputProps, shouldRenderSuggestions } = props
    const { value } = inputProps

    return suggestions.length > 0 && shouldRenderSuggestions(value)
  }

  storeReferences = autowhatever => {
    if (autowhatever !== null) {
      const { input, itemsContainer } = autowhatever

      this.input = input
      this.suggestionsContainer = itemsContainer
    }
  }

  onSuggestionMouseEnter = (event, { sectionIndex, itemIndex }) => {
    this.updateFocusedSuggestion(sectionIndex, itemIndex)
  }

  focusFirstSuggestion = () => {
    this.updateFocusedSuggestion(this.props.multiSection ? 0 : null, 0)
  }

  onSuggestionMouseDown = () => {
    this.justSelectedSuggestion = true
  }

  onSuggestionsClearRequested = () => {
    const { onSuggestionsClearRequested } = this.props

    onSuggestionsClearRequested && onSuggestionsClearRequested()
  }

  onSuggestionSelected = (event, data) => {
    const {
      alwaysRenderSuggestions,
      onSuggestionSelected,
      onSuggestionsFetchRequested
    } = this.props

    onSuggestionSelected && onSuggestionSelected(event, data)

    if (alwaysRenderSuggestions) {
      onSuggestionsFetchRequested({ value: data.suggestionValue })
    } else {
      this.onSuggestionsClearRequested()
    }

    this.resetFocusedSuggestion()
  }

  onSuggestionClick = event => {
    const { alwaysRenderSuggestions, focusInputOnSuggestionClick } = this.props
    const { sectionIndex, suggestionIndex } = this.getSuggestionIndices(
      this.findSuggestionElement(event.target)
    )
    const clickedSuggestion = this.getSuggestion(sectionIndex, suggestionIndex)
    const clickedSuggestionValue = this.props.getSuggestionValue(
      clickedSuggestion
    )

    this.maybeCallOnChange(event, clickedSuggestionValue, "click")
    this.onSuggestionSelected(event, {
      suggestion: clickedSuggestion,
      suggestionValue: clickedSuggestionValue,
      suggestionIndex: suggestionIndex,
      sectionIndex,
      method: "click"
    })

    if (!alwaysRenderSuggestions) {
      this.closeSuggestions()
    }

    if (focusInputOnSuggestionClick === true) {
      this.input.focus()
    } else {
      this.onBlur()
    }

    setTimeout(() => {
      this.justSelectedSuggestion = false
    })
  }

  onBlur = () => {
    const { inputProps, shouldRenderSuggestions } = this.props
    const { value, onBlur } = inputProps
    const focusedSuggestion = this.getFocusedSuggestion()

    this.inputBlurred(shouldRenderSuggestions(value))
    onBlur && onBlur(this.blurEvent, { focusedSuggestion })
  }

  resetFocusedSuggestionOnMouseLeave = () => {
    this.resetFocusedSuggestion(false) // shouldResetValueBeforeUpDown
  }

  itemProps = ({ sectionIndex, itemIndex }) => {
    return {
      "data-section-index": sectionIndex,
      "data-suggestion-index": itemIndex,
      onMouseEnter: this.onSuggestionMouseEnter,
      onMouseLeave: this.resetFocusedSuggestionOnMouseLeave,
      onMouseDown: this.onSuggestionMouseDown,
      onTouchStart: this.onSuggestionMouseDown, // Because on iOS `onMouseDown` is not triggered
      onClick: this.onSuggestionClick
    }
  }

  render() {
    const {
      suggestions,
      renderInputComponent,
      renderSuggestionsContainer,
      onSuggestionsFetchRequested,
      renderSuggestion,
      inputProps,
      multiSection,
      renderSectionTitle,
      id,
      getSectionSuggestions,
      getSuggestionValue,
      alwaysRenderSuggestions
    } = this.props
    const {
      isFocused,
      isCollapsed,
      focusedSectionIndex,
      focusedSuggestionIndex,
      valueBeforeUpDown
    } = this.state
    const shouldRenderSuggestions = alwaysRenderSuggestions
      ? alwaysTrue
      : this.props.shouldRenderSuggestions
    const { value, onFocus, onKeyDown } = inputProps
    const willRenderSuggestions = this.willRenderSuggestions(this.props)
    const isOpen =
      alwaysRenderSuggestions ||
      (isFocused && !isCollapsed && willRenderSuggestions)
    const items = isOpen ? suggestions : []
    const autowhateverInputProps = {
      ...inputProps,
      onFocus: event => {
        if (
          !this.justSelectedSuggestion &&
          !this.justClickedOnSuggestionsContainer
        ) {
          const shouldRender = shouldRenderSuggestions(value)

          this.inputFocused(shouldRender)
          onFocus && onFocus(event)

          if (shouldRender) {
            onSuggestionsFetchRequested({ value })
          }
        }
      },
      onBlur: event => {
        if (this.justClickedOnSuggestionsContainer) {
          this.input.focus()
          return
        }

        this.blurEvent = event

        if (!this.justSelectedSuggestion) {
          this.onBlur()
          this.onSuggestionsClearRequested()
        }
      },
      onChange: event => {
        const { value } = event.target
        const shouldRender = shouldRenderSuggestions(value)

        this.maybeCallOnChange(event, value, "type")
        this.inputChanged(shouldRender)

        if (shouldRender) {
          onSuggestionsFetchRequested({ value })
        } else {
          this.onSuggestionsClearRequested()
        }
      },
      onKeyDown: (event, data) => {
        switch (event.key) {
          case "ArrowDown":
          case "ArrowUp":
            if (isCollapsed) {
              if (shouldRenderSuggestions(value)) {
                onSuggestionsFetchRequested({ value })
                this.revealSuggestions()
              }
            } else if (suggestions.length > 0) {
              const { newFocusedSectionIndex, newFocusedItemIndex } = data

              let newValue

              if (newFocusedItemIndex === null) {
                // valueBeforeUpDown can be null if, for example, user
                // hovers on the first suggestion and then pressed Up.
                // If that happens, use the original input value.
                newValue = valueBeforeUpDown === null
                  ? value
                  : valueBeforeUpDown
              } else {
                newValue = this.getSuggestionValueByIndex(
                  newFocusedSectionIndex,
                  newFocusedItemIndex
                )
              }

              this.updateFocusedSuggestion(
                newFocusedSectionIndex,
                newFocusedItemIndex,
                value
              )
              this.maybeCallOnChange(
                event,
                newValue,
                event.key === "ArrowDown" ? "down" : "up"
              )
            }

            event.preventDefault() // Prevents the cursor from moving

            this.justPressedUpDown = true

            setTimeout(() => {
              this.justPressedUpDown = false
            })

            break

          case "Enter": {
            const focusedSuggestion = this.getFocusedSuggestion()

            if (isOpen && !alwaysRenderSuggestions) {
              this.closeSuggestions()
            }

            if (focusedSuggestion !== null) {
              const newValue = getSuggestionValue(focusedSuggestion)

              this.maybeCallOnChange(event, newValue, "enter")

              this.onSuggestionSelected(event, {
                suggestion: focusedSuggestion,
                suggestionValue: newValue,
                suggestionIndex: focusedSuggestionIndex,
                sectionIndex: focusedSectionIndex,
                method: "enter"
              })

              this.justSelectedSuggestion = true

              setTimeout(() => {
                this.justSelectedSuggestion = false
              })
            }

            break
          }

          case "Escape": {
            if (isOpen) {
              // If input.type === 'search', the browser clears the input
              // when Escape is pressed. We want to disable this default
              // behaviour so that, when suggestions are shown, we just hide
              // them, without clearing the input.
              event.preventDefault()
            }

            const willCloseSuggestions = isOpen && !alwaysRenderSuggestions

            if (valueBeforeUpDown === null) {
              // Didn't interact with Up/Down
              if (!willCloseSuggestions) {
                const newValue = ""

                this.maybeCallOnChange(event, newValue, "escape")

                if (shouldRenderSuggestions(newValue)) {
                  onSuggestionsFetchRequested({ value: newValue })
                } else {
                  this.onSuggestionsClearRequested()
                }
              }
            } else {
              // Interacted with Up/Down
              this.maybeCallOnChange(event, valueBeforeUpDown, "escape")
            }

            if (willCloseSuggestions) {
              this.onSuggestionsClearRequested()
              this.closeSuggestions()
            } else {
              this.resetFocusedSuggestion()
            }

            break
          }
        }

        onKeyDown && onKeyDown(event)
      }
    }
    const renderSuggestionData = {
      query: (valueBeforeUpDown || value).trim()
    }

    return (
      <Autowhatever
        multiSection={multiSection}
        items={items}
        renderInputComponent={renderInputComponent}
        renderItemsContainer={renderSuggestionsContainer}
        renderItem={renderSuggestion}
        renderItemData={renderSuggestionData}
        renderSectionTitle={renderSectionTitle}
        getSectionItems={getSectionSuggestions}
        focusedSectionIndex={focusedSectionIndex}
        focusedItemIndex={focusedSuggestionIndex}
        inputProps={autowhateverInputProps}
        itemProps={this.itemProps}
        theme={defaultTheme}
        id={id}
        ref={this.storeReferences}
      />
    )
  }
}
